/**
 * @file metadataHolder.h
 *
 * @ingroup iopccommon
 * @author Petr Cermak
 */


#ifndef METADATAHOLDER_H_
#define METADATAHOLDER_H_

#include <string>
#include <map>
#include <iostream>
#include "trace.h"
#include "exceptionBase.h"

namespace iopc {

/**
 * @brief A metadata entry
 *
 * Instance of this class represents a single metadata value.
 * Values of the data types listed in @a Metadata::MetadataType
 * can be represented by instances of this class.
 * This class can be used as a value type.
 *
 * @ingroup iopccommon
 *
 * The class is not thread-safe.
 */
class Metadata {
public:

	/**
	 * @brief A metadata entry type
	 */
	enum MetadataType {
		INT, 	 /**< Integer value */
		STRING,  /**< String value */
		BOOL,	 /**< Boolean value */
		POINTER, /**< A void* value */
		NOTHING  /**< This type represents empty metadata */
	};

	/**
	 * @brief Initializes the metadata to the @a NOTHING @a MetadataType
	 */
	Metadata() {
		type = NOTHING;
		intValue = 0;
		pointerValue = NULL;
		boolValue = false;
	}
	/**
	 * @brief Initializes the metadata with an integer value
	 */
	Metadata(int value) {
		setIntValue(value);
	}
	/**
	 * @brief Initializes the metadata with a string value
	 */
	Metadata(const std::string& value) {
		setStringValue(value);
	}
	/**
	 * @brief Initializes the metadata with a string value
	 */
	Metadata(const char* value) {
		setStringValue(value);
	}
	/**
	 * @brief Initializes the metadata with a boolean value
	 */
	Metadata(bool value) {
		setBoolValue(value);
	}
	/**
	 * @brief Initializes the metadata with a pointer value
	 */
	Metadata(void* value) {
		setPointerValue(value);
	}

	/**
	 * @brief Changes the value and type of the metadata
	 */
	inline void setIntValue(int value) {
		type = INT;
		intValue = value;
	}

	/**
	 * @brief  Changes the value and type of the metadata
	 */
	inline void setStringValue(const std::string& value) {
		type = STRING;
		stringValue = value;
	}
	/*inline void setStringValue(const char* value) {
		type = STRING;
		stringValue = value;
	}*/
	/**
	 * @brief  Changes the value and type of the metadata
	 */
	inline void setBoolValue(bool value) {
		type = BOOL;
		boolValue = value;
	}
	/**
	 * @brief  Changes the value and type of the metadata
	 */
	inline void setPointerValue(void* value) {
		type = POINTER;
		pointerValue = value;
	}

	/**
	 * @brief Returns the metadata value.
	 *
	 * @throws IopcIncorrectUsageException if the metadata is not of the @a Metadata::INT type
	 */
	inline int getIntValue() const {
		if (!type == INT) {
			IOPC_LOG_THROW(IopcIncorrectUsageException("Wrong type of metadata"));
		}
		return intValue;
	}

	/**
	 * @brief Returns the metadata value.
	 *
	 * @throws IopcIncorrectUsageException if the metadata is not of the @a Metadata::STRING type
	 */
	inline std::string getStringValue() const {
		if (!type == STRING) {
			IOPC_LOG_THROW(IopcIncorrectUsageException("Wrong type of metadata"));
		}
		return stringValue;
	}

	/**
	 * @brief Returns the metadata value.
	 *
	 * @throws IopcIncorrectUsageException if the metadata is not of the @a Metadata::BOOL type
	 */
	inline bool getBoolValue() const {
		if (!type == BOOL) {
			IOPC_LOG_THROW(IopcIncorrectUsageException("Wrong type of metadata"));
		}
		return boolValue;
	}

	/**
	 * @brief Returns the metadata value.
	 *
	 * @throws IopcIncorrectUsageException if the metadata is not of the @a Metadata::POINTER type
	 */
	inline void* getPointerValue() const {
		if (!type == POINTER) {
			IOPC_LOG_THROW(IopcIncorrectUsageException("Wrong type of metadata"));
		}
		return pointerValue;
	}

	operator const bool() const {
		return getBoolValue();
	}
	operator const int() const {
		return getIntValue();
	}
	operator const std::string() const {
		return getStringValue();
	}
	/*operator const char*() const {
		return getStringValue().c_str();
	}*/
	operator const void*() const {
		return getPointerValue();
	}

	/**
	 * @brief Returns true if the metadata type is @a Metadata::NOTHING
	 */
	inline bool isEmpty() const {
		return type == NOTHING;
	}

	// only the NOTHING type can be compared at the moment...
	bool operator==(const Metadata& src) {
		return (src.type == NOTHING) && (this->type == NOTHING);
	}

	/**
	 * @brief Returns a string representation of the metadata value.
	 */
	std::string toString() const;

	friend std::ostream& operator<<(std::ostream& out, const Metadata& m);

private:
	int intValue;
	bool boolValue;
	void* pointerValue;
	std::string stringValue;
	MetadataType type;
};

/**
 * Outputs a string representation of the metadata type and value into a stream.
 */
std::ostream& operator<<(std::ostream& out, const Metadata& m);

/**
 * @brief A dictionary-like container class used for class metadata storage.
 *
 * MetadataHolder can be used in a similar way as STL map container.
 * It overrides the [] operator. Keys are string values, items are @a Metadata objects.
 *
 * Other @a MetadataHolder instance can be attached using @a setSecondaryMetadataLocation()
 * to provide a secondary search location for meatadata key lookup.
 *
 * A global @a MetadataHolder instance called 'defaults' is searched if the requested metadata
 * is not found in the current instance or in the secondary metadata location (if specified).
 * This @a MetadataHolder instance can be obtained by calling the @a getDefaultsMeta() method.
 *
 * If a metadata is requested, the default search order is:
 * <ol>
 * 	<li>Current @a MetadataHolder instance</li>
 *  <li>Secondary metadata location specified by the @a setSecondaryMetadataLocation() method</li>
 *  <li>Default metadata location in @a getDefaultsMeta()
 * </ol>
 *
 * Usage:
 * @code
 * mholder[key] = "stringVal";
 * mholder[key].setStringValue("stringVal");
 * stringVal = mholder[key];
 * stringVal = mholder[key].getStringValue();
 * @endcode
 *
 * The class is not thread-safe.
 *
 * @ingroup iopccommon
 * @see Metadata
 */
class MetadataHolder {
private:
	// The metadata container
	mutable std::map<std::string, Metadata> meta;
	const MetadataHolder* secondaryMeta;
public:
	MetadataHolder();

	/**
	 * @brief Returns the solitary default @a MetadataHolder instance
	 */
	static MetadataHolder& getDefaultsMeta();

	/**
	 * @brief Asks the default @a MetadataHolder instance if it contains metadata of the specified key
	 */
	static inline bool hasDefaultMeta(const std::string& key) {
		return getDefaultsMeta().hasMeta(key);
	}

	/**
	 * @brief Returns @a Metadata of the specified key
	 *
	 * @param key Key of the requested metadata
	 * @param searchDefaults Allows to skip the default metadata search
	 * @param searchSecondary Allows to skip the secondary location metadata search
	 *
	 * @returns A @a Metadata instance found under the specified key
	 *
	 * @throws IopcIncorrectUsageException if no metadata of such key is found
	 */
	virtual const Metadata& getMeta(const std::string& key, bool searchDefaults = true, bool searchSecondary = true) const;

	/**
	 * @brief Sets metadata for a specified key
	 *
	 * If specified metadata already exist, the metadata value is overwritten.
	 *
	 * @param key Metadata key
	 * @param value New metadata value
	 */
	void setMeta(const std::string& key, const Metadata& value) const;

	/**
	 * @brief Sets a secondary search location.
	 *
	 * See the @a MetadataHolder description
	 */
	void setSecondaryMetadataLocation(const MetadataHolder* secondaryMeta);

	/**
	 * @brief Asks the @a MetadataHolder if it contains metadata of the specified key
	 *
	 * @param key Key of the requested metadata
	 * @param searchDefaults Allows to skip the default metadata search
	 * @param searchSecondary Allows to skip the secondary location metadata search
	 */
	virtual bool hasMeta(const std::string& key, bool searchDefaults = true, bool searchSecondary = true) const;
	// hasMeta && getBoolValue == true

	/**
	 * @brief Shortcut for @a hasMeta(key) and @a getMeta(key).getBoolValue()
	 */
	inline bool hasFeature(const std::string& key) const {
		return hasMeta(key) && getMeta(key).getBoolValue();
	}

	/**
	 * @brief Shortcut for setting a boolean metadata (true) under the specified key
	 */
	inline void setFeature(const std::string& key) const {
		meta[key].setBoolValue(true);
	}

	/**
	 * @brief Asks the @a MetadataHolder if the secondary metadata location was specified
	 */
	inline bool hasSecondaryMetadataLocation() const {
		return secondaryMeta != NULL;
	}

	/**
	 * @brief A dictionary-like access to the metadata
	 *
	 * This operator doesn't access the default values and secondary metadata location
	 * when looking up the metadata value!
	 */
	inline Metadata& operator[](const std::string& key) const {
		return meta[key];
	}

	static MetadataHolder emptyMeta;

	virtual ~MetadataHolder() {};

	inline bool operator!=(const MetadataHolder& src) const { return this != &src; }
	inline bool operator==(const MetadataHolder& src) const { return this == &src; }
	friend std::ostream& operator <<(std::ostream& out, const MetadataHolder& m);

	/**
	 * @brief Returns a string representation of the metadata value.
	 *
	 * Lists all @a Metadata::INT, @a Metadata::STRING and @a Metadata::BOOL values
	 */
	std::string toString() const;
};
std::ostream& operator<<(std::ostream& out, const MetadataHolder& m);

/**
 * @brief A @a MetadataHolder wrapper
 *
 * This class can be used to specify additional metadata to an existing @a MetadataHolder instance
 * if we don't want to modify its contents.
 *
 * @ingroup iopccommon
 */
class AdditionalMetadataHolder : public MetadataHolder
{
private:
	const MetadataHolder& primary;
	bool invert;
public:
	/**
	 * @param primary The @a MetadataHolder instance being wrapped
	 * @param invert Specifies the search order.
	 * 	If invert is false (by default), the wrapped instance is searched first,
	 *  otherwise this wrapper is searched before the wrapped @a MetadataHolder.
	 */
	AdditionalMetadataHolder(const MetadataHolder& primary, bool invert = false) : primary(primary), invert(invert) {};

	/**
	 * @brief Returns @a Metadata of the specified key
	 *
	 * @param key Key of the requested metadata
	 * @param searchDefaults Allows to skip the default metadata search
	 * @param searchSecondary Allows to skip the secondary location of both
	 * 	this wrapper and the wrapped @a MetadataHolder instance
	 *
	 * @returns A @a Metadata instance found under the specified key.
	 *
	 * @throws IopcIncorrectUsageException if no metadata of such key is found
	 */
	virtual const Metadata& getMeta(const std::string& key, bool searchDefaults = true, bool searchSecondary = true) const;

	/**
	 * @brief Asks the wrapper and the wrapped @a MetadataHolder instance if they contain metadata of the specified key
	 *
	 * @param key Key of the requested metadata
	 * @param searchDefaults Allows to skip the default metadata search
	 * @param searchSecondary Allows to skip the secondary location metadata search
	 *
	 * @return true - this wrapper or the wrapped @a MetadtaHolder instance contain the key
	 */
	virtual bool hasMeta(const std::string& key, bool searchDefaults = true, bool searchSecondary = true) const;
};

}
#endif /*METADATAHOLDER_H_*/
