/**
 * @file connection.h
 *
 * @author Michal Kopecky
 * @author Petr Cermak
 *
 * @ingroup iopcdb
 */

#ifndef CONNECTION_H_
#define CONNECTION_H_

#include <string>
#include "iopccommon/defs.h"


namespace iopc {

class Cursor;
class Database;
class ConnectionImpl;
class Driver;

/**
 * @brief Represents a connection to a SQL database.
 *
 * This class encapsulates a single connection to a database.
 * @a Connection  allows users to send SQL commands to its associated database and
 * to control transaction processing.
 * It manages the cursors that are created by executing
 * SQL via @a sql(const std::string& name) or getCursor(const std::string& name) methods.
 *
 * @ingroup iopcdb
 */
class DLL_EXTERNAL Connection
{
public:
	virtual ~Connection() {};

	/**
	 * @brief Sends an SQL statement (a query) to an associated database.
	 *
	 * @returns A @a Cursor object representing the result set of the SQL command
	 */
	virtual Cursor* sql(const std::string& name) = 0;

	/**
	 * @brief Cursors that are no longer needed should be returned to the driver @a Connection object
	 * using this method.
	 *
	 * All resources allocated by the @a Cursor are released.
	 */
	virtual void returnCursor(Cursor* cur) = 0;

	/**
	 * @brief Closes and returns all cursors on this connection.
	 */
	virtual void closeAllCursors() = 0;

	/**
	 * @brief Executes an SQL statement that does not return any rows (DDL, DML).
	 *
	 * No @a Cursor object is created
	 */
	virtual void sqlNonQuery(const std::string& name) = 0;

	/**
	 * @brief Commits current transaction. Starts new one.
	 */
	virtual void commit() = 0;

	/**
	 * @brief Rolls back current transaction.
	 */
	virtual void rollback() = 0;

	/**
	 * @brief Creates a savepoint
	 *
	 * @param name Savepoint name
	 */
	virtual void savepoint(const std::string& name) = 0;

	/**
	 * @brief Rolls back to the specified savepoint
	 *
	 * @param name Name of the savepoint to which to roll back to.
	 */
	virtual void rollbackToSavepoint(const std::string& name) = 0;

	/**
	 * @brief Opens the connection
	 *
	 * This call opens a physical link to the DBMS
	 */
	virtual void open() = 0;

	/**
	 * @brief Closes the connection
	 *
	 * This call should close the physical link to the DBMS
	 */
	virtual void close() = 0;


	/**
	 *  @brief Commits last transaction and closes the connection.
	 *
	 *  The connection still needs to be returned to associated @a Database object
	 */
	virtual void disconnect() = 0;

	/**
	 * @brief Rolls back current transaction and closes the connection
	 */
	virtual void abort() = 0;

	/**
	 * @brief Turns autocommit on this connection on or off
	 *
	 * @param ac - true - autocommit is enabled, false - autocommit is disabled
	 */
	virtual void setAutoCommit(bool ac) = 0;

	/**
	 *  @brief Determines whether this connection is opened or not
	 */
	virtual bool isOpened() = 0;

	/**
	 *  @brief Shortcut for @a sqlNonQuery
	 */
	virtual Connection& operator<< (
	    const std::string& command
	) = 0;

	virtual bool operator== (const Connection& dbConn) const = 0;

	virtual bool operator!= (const Connection& dbConn) const = 0;

	/**
	 * @brief Returns the associated @a Database object
	 */
	virtual Database& getDatabase() = 0;

	/**
	 * @brief Returns the database driver used
	 */
	virtual Driver& getDriver() = 0;

	/**
	 * @brief Returns the driver-side part of the @a Connection implemenation.
	 */
	virtual ConnectionImpl* getImplementation() = 0;

	/**
	 * @brief Returns the driver-independent part of the @a Connection implementation.
	 */
	virtual Connection* getBasicConnection() = 0;
};

}

#endif /*CONNECTION_H_*/
