/**
 * @file cursor.h
 *
 * @author Michal Kopecky
 * @author Josef Troch Josef Troch (<a href="mailto:josef.troch@email.cz">josef.troch@email.cz</a>)
 * @author Jan Hadrava
 * @author Petr Cermak
 *
 * @ingroup iopcdb
 */

#ifndef CURSOR_H_
#define CURSOR_H_

#include "connection.h"
#include "iopccommon/defsPlatform.h"
#include "cursorImpl.h"
#include "iopcmeta/type.h"
#include "iopcmeta/enhancedTypeBase.h"
#include "iopcmeta/enhancedString.h"

#include <string>
#include <map>

namespace iopc {

/**
 * @brief A parameter that cen be passed to a cursor.
 *
 * @author Petr Cermak
 *
 * @ingroup iopcdb
 */
class Parameter
{
public:
	/**
	 * @brief Parameter direction
	 */
	enum ParamType {
		/**
		 * @brief IN parameter
		 *
		 * Its implementations are notified with the @a preExec and @a postExec event methods
		 */
		IN = 0,

		/**
		 * @brief OUT parameter
		 *
		 * Its implementations are notified with the @a preFetch and @a postFetch event methods
		 */
		OUT
	};

	/**
	 * @param type Parameter direction
	 * @param pos Position of the parameter in the prepared statement or in the result set
	 * @param varType Parameter data type
	 * @param attributes Additional attributes or driver-dependent data
	 * @param setNull Determines whether this parameter should be set to database NULL (for IN parameters)
	 */
	Parameter(ParamType type, const int pos, const Type& varType, const MetadataHolder& attributes = MetadataHolder::emptyMeta, bool setNull = false);
	virtual ~Parameter() {}

	/**
	 * @brief Notifies the parameter before @a Cursor::execute() is invoked
	 */
	void preExec(Cursor* cur);

	/**
	 * @brief Notifies the parameter after @a Cursor::execute() call
	 */
	void postExec(Cursor* cur);
	/**
	 * @brief Notifies the parameter before first @a Cursor::fetchNext() after @a Cursor::execute() is invoked
	 */
	void preFetch(Cursor* cur);
	/**
	 * @brief Notifies the parameter after each @a Cursor::fetchNext() call
	 */
	void postFetch(Cursor* cur);

	/**
	 * @brief Determines if the last fetched value was database NULL
	 */
	inline bool isNullValue() { return isNull; }

protected:
	virtual void doPreExec(Cursor* cur) {};
	virtual void doPostExec(Cursor* cur) {};
	virtual void doPreFetch(Cursor* cur) {};
	virtual void doPostFetch(Cursor* cur) {};

	ParamType paramType;
	const int pos;
	const Type& varType;
	const MetadataHolder& attributes;
	bool setNull;
	bool isNull;
};

/**
 * @brief A @a Parameter implementation for basic C++ numeric and string types.
 *
 * Requires the @a varType to be of Type::IOPC_TYPECLASS_STRING or Type::IOPC_TYPECLASS_SIMPLE
 *
 * @ingroup iopcdb
 */
class SimpleTypeParameter : public Parameter
{
public:

	/**
	 * @param type Parameter direction
	 * @param pos Position of the parameter in the prepared statement or in the result set, starting from 1.
	 * 	Note that some drivers (like the IopcOracle10g driver) require the OUT parameters to be bound as a sequence so that
	 *  no positions are skipped.
	 * @param varAddr Address of the input parameter value or an output buffer to be bound
	 * 	If Type::IOPC_TYPECLASS_STRING type is used, this pointer must point to a buffer containing
	 *  standard, zero-terminated C string representation [IN] or to a buffer into which a standard C string
	 *  will be fetched [OUT]. So even if the varType is TypeDesc<std::string>
	 *  the varAddr must point [IN] or will point [OUT] to a standard C string.
	 * @param varType Parameter data type
	 * @param attributes Additional attributes or driver-dependent data.
	 * <ul>
	 * <li>[db.type.length] - For the Type::IOPC_TYPECLASS_STRING types it specifies the
	 *     maximum allocated length (in characters without the trailing zero) for the input and output buffers.</li>
	 * </ul>
	 *
	 * If Type::IOPC_TYPECLASS_STRING type is used, the [db.type.length] metadata must be specified either in the @a varType metadata or
	 * provided in the @a attributes input parameter.
	 */
	SimpleTypeParameter(ParamType type, const int pos, const void* varAddr, const Type& varType, const MetadataHolder& attributes = MetadataHolder::emptyMeta);

protected:
	virtual void doPreExec(Cursor* cur);
	virtual void doPreFetch(Cursor* cur);
	virtual void doPostFetch(Cursor* cur);

private:
	const void* varAddr;
	unsigned int maxLen;
	EString tmpStr;
	EWString tmpWStr;

};

/**
 * @brief A @a Parameter implementation for all Enhanced data types.
 *
 * It provides more comfort than @a SimpleTypeParameter as users need not to provide any
 * pointers, any data type descriptions or allocate any buffers. All these tasks are handled by the Enhanced data types.
 *
 * @ingroup iopcdb
 */
class EnhancedTypeParameter : public Parameter
{
public:

	/**
	 * @param type Parameter direction
	 * @param pos Position of the parameter in the prepared statement or in the result set, starting from 1.
	 * 	Note that some drivers (like the IopcOracle10g driver) require the OUT parameters to be bound as a sequence so that
	 *  no positions are skipped.
	 * @param obj Reference to the value being bound. The reference must be valid as long as the cursor is used.
	 * @param attributes Additional attributes or driver-dependent data.
	 * <ul>
	 * <li>[db.type.length] - For the @a EString and @a EWString types it specifies the
	 *     (maximum) length of the externalized string buffer - see the externalize methods in @a EnhancedTypeBase.</li>
	 * </ul>
	 *
	 * If @a EString or @a EWString type is used, the [db.type.length] metadata must be specified either
	 * in the TypeDesc<EString>::getType() or in the TypeDesc<EWString>::getType metadata or
	 * provided in the @a attributes input parameter.
	 */
	EnhancedTypeParameter(ParamType type, const int pos, EnhancedTypeBase& obj, const MetadataHolder& attributes = MetadataHolder::emptyMeta);
protected:
	virtual void doPreExec(Cursor* cur);
	virtual void doPreFetch(Cursor* cur);
	virtual void doPostFetch(Cursor* cur);

private:
	EnhancedTypeBase& variable;
	unsigned int maxLen;
};

/**
 * @brief A @a Parameter implementation representing a NULL INput parameter.
 *
 * Currently only the IN direction is supported.
 *
 * @ingroup iopcdb
 */
class NullParameter : public Parameter
{
public:

	/**
	 * @param type Parameter direction. Only the IN direction is supported, if OUT is specified, the behaviour is undefined.
	 * @param pos Position of the parameter in the prepared statement or in the result set, starting from 1.
	 * 	Note that some drivers (like the IopcOracle10g driver) require the parameters to be bound as a sequence so that
	 *  no positions are skipped.
	 * @param varType Data type of the parameter. Enhanced as well as simple types may be used.
	 * @param attributes Additional attributes or driver-dependent data.
	 */
	NullParameter(ParamType type, const int pos, const Type& varType, const MetadataHolder& attributes = MetadataHolder::emptyMeta);

protected:
	virtual void doPreExec(Cursor* cur);

private:
	void* value;
};

/**
 * @brief Executes SQL statements and fetches the results.
 *
 * Also allows user to bind input parameters to parametrised queries and
 * output parameters into which the execution results will be fetched.
 *
 * This class is NOT thread safe and is not intended to be used concurrently.
 *
 * @ingroup iopcdb
 */
class DLL_EXTERNAL Cursor
{
public:
	/**
	 * @brief State of the cursor
	 *
	 * Cursor is created as @a CLOSED,
	 * The @a open() method changes its state to @a Cursor::OPENED,
	 * @a prepare() method changes its state to @a Cursor::PREPARED
	 * and @a execute() to @a Cursor::EXECUTED.
	 * Then the cursor can be @a Cursor::PREPARED again or
	 * @a Cursor::CLOSED.
	 */
	enum State {
		CLOSED, OPENED, PREPARED, EXECUTED
	};

private:
	State state;
	CursorImpl* implementation;
	std::string sql;

	typedef std::map<int, Parameter*> ParametersMap;
	typedef ParametersMap::const_iterator ParametersIterator;

	ParametersMap inParams;
	ParametersMap outParams;

	bool fetched;

	void prepareState();
	void checkAndDeleteParam(const int pos, ParametersMap& map);
public:
	/**
	 * @brief Adds a new IN parameter. Accepts Type::IOPC_TYPECLASS_STRING, Type::IOPC_TYPECLASS_SIMPLE as well as
	 * Type::IOPC_TYPECLASS_ENHANCED.
	 *
	 * If the @a varType is of @a Type::IOPC_TYPECLASS_ENHANCED, the @a EnhancedTypeParameter is created.
	 * @a SimpleTypeParameter is created for @a Type::IOPC_TYPECLASS_SIMPLE and @a Type::IOPC_TYPECLASS_ENHANCED
	 * type classes.
	 *
	 * If string types (Type::IOPC_TYPECLASS_STRING, EString or EWString), the [db.type.length] metadata must be specified either
	 * in the @a varType metadata or provided in the @a attributes input parameter.
	 *
	 * @throws IopcIncorrectUsageException if other type than @a Type::IOPC_TYPECLASS_ENHANCED, @a Type::IOPC_TYPECLASS_SIMPLE
	 * or @a Type::IOPC_TYPECLASS_ENHANCED  is specified.
	 *
	 * @param pos Position of the parameter in the prepared statement, starting from 1.
	 * 	If a parameter already exists on the @a pos position, it is overwritten.
	 * @param varAddr Pointer to the value being bound. The pointer must remain valid as long as the cursor is used.
	 * 	If Type::IOPC_TYPECLASS_STRING type is used, this pointer must point to a buffer containing
	 *  standard, zero-terminated C string representation. So even if the varType is TypeDesc<std::string>
	 *  the varAddr must point to a standard C string.
	 *  If EString or EWString is used, the pointer should point to their instances.
	 * @param varType Parameter data type
	 * @param attributes Additional attributes or driver-dependent data.
	 * <ul>
	 * <li>[db.type.length] - For the @a EString and @a EWString types it specifies the
	 *     (maximum) length of the externalized string buffer - see the externalize methods in @a EnhancedTypeBase.
	 *     For the Type::IOPC_TYPECLASS_STRING types it specifies the
	 *     maximum allocated length (in characters without the trailing zero) for the input buffers.</li>
	 * </ul>
	 *
	 */
	void addInParam(const int pos, const void* varAddr, const Type& varType, const MetadataHolder& attributes = MetadataHolder::emptyMeta);

	/**
	 * @brief Adds a new Enhanced data type IN parameter
	 *
	 * @param pos Position of the parameter in the prepared statement, starting from 1.
	 * 	If a parameter already exists on the @a pos position, it is overwritten.
	 * @param obj Reference to the value being bound. The reference must be valid as long as the cursor is used.
	 * @param attributes Additional attributes or driver-dependent data.
	 * <ul>
	 * <li>[db.type.length] - For the @a EString and @a EWString types it specifies the
	 *     (maximum) length of the externalized string buffer - see the externalize methods in @a EnhancedTypeBase.</li>
	 * </ul>
	 *
	 * If @a EString or @a EWString type is used, the [db.type.length] metadata must be specified either
	 * in the TypeDesc<EString>::getType() or in the TypeDesc<EWString>::getType metadata or
	 * provided in the @a attributes input parameter.
	 */
	void addEnhancedInParam(const int pos, EnhancedTypeBase& obj, const MetadataHolder& attributes = MetadataHolder::emptyMeta);

	/**
	 * @brief Adds an database NULL IN parameter
	 *
	 * @param pos Position of the parameter in the prepared statement or in the result set, starting from 1.
	 * 	Note that some drivers (like the IopcOracle10g driver) require the parameters to be bound as a sequence so that
	 *  no positions are skipped.
	 * @param type Data type of the parameter. Enhanced as well as simple types may be used.
	 * @param attributes Additional attributes or driver-dependent data.
	 */
	void addNullInParam(const int pos, const Type& type, const MetadataHolder& attributes = MetadataHolder::emptyMeta);

	/**
	 * @brief Adds a new OUT parameter. Accepts Type::IOPC_TYPECLASS_STRING, Type::IOPC_TYPECLASS_SIMPLE as weel as
	 * Type::IOPC_TYPECLASS_ENHANCED.
	 *
	 * If the @a varType is of @a Type::IOPC_TYPECLASS_ENHANCED, the @a EnhancedTypeParameter is created.
	 * @a SimpleTypeParameter is created for @a Type::IOPC_TYPECLASS_SIMPLE and @a Type::IOPC_TYPECLASS_ENHANCED
	 * type classes.
	 *
	 * If string types (Type::IOPC_TYPECLASS_STRING, EString or EWString), the [db.type.length] metadata must be specified either
	 * in the @a varType metadata or provided in the @a attributes input parameter.
	 *
	 * @throws IopcIncorrectUsageException if other type than @a Type::IOPC_TYPECLASS_ENHANCED, @a Type::IOPC_TYPECLASS_SIMPLE
	 * or @a Type::IOPC_TYPECLASS_ENHANCED  is specified.
	 *
	 * @param pos Position of the parameter in the prepared statement or in the result set, starting from 1.
	 * 	Note that some drivers (like the IopcOracle10g driver) require the OUT parameters to be bound as a sequence so that
	 *  no positions are skipped. If a parameter already exists on the @a pos position, it is overwritten.
	 * @param varAddr Pointer to the value being bound. The pointer must remain valid as long as the cursor is used.
	 * 	If Type::IOPC_TYPECLASS_STRING type is used, this pointer must point to a buffer into which a standard C string
	 *  will be fetched. So even if the varType is TypeDesc<std::string>
	 *  the varAddr will point to a standard C string.
	 *  If EString or EWString is used, the pointer should point to their instances.
	 * @param varType Parameter data type
	 * @param attributes Additional attributes or driver-dependent data.
	 * <ul>
	 * <li>[db.type.length] - For the @a EString and @a EWString types it specifies the
	 *     (maximum) length of the externalized string buffer - see the externalize methods in @a EnhancedTypeBase.
	 *     For the Type::IOPC_TYPECLASS_STRING types it specifies the
	 *     maximum allocated length (in characters without the trailing zero) for the output buffers.</li>
	 * </ul>
	 *
	 */
	void addOutParam(const int pos, const void* varAddr, const Type& varType, const MetadataHolder& attributes = MetadataHolder::emptyMeta);

	/**
	 * @brief Adds a new Enhanced data type OUT parameter
	 *
	 * @param pos Position of the parameter in the prepared statement, starting from 1.
	 * 	If a parameter already exists on the @a pos position, it is overwritten.
	 * @param obj Reference to the value being bound. The reference must be valid as long as the cursor is used.
	 * @param attributes Additional attributes or driver-dependent data.
	 * <ul>
	 * <li>[db.type.length] - For the @a EString and @a EWString types it specifies the
	 *     (maximum) length of the externalized string buffer - see the externalize methods in @a EnhancedTypeBase.</li>
	 * </ul>
	 *
	 * If @a EString or @a EWString type is used, the [db.type.length] metadata must be specified either
	 * in the TypeDesc<EString>::getType() or in the TypeDesc<EWString>::getType metadata or
	 * provided in the @a attributes input parameter.
	 */
	void addEnhancedOutParam(const int pos, EnhancedTypeBase& obj, const MetadataHolder& attributes = MetadataHolder::emptyMeta);

	/**
	 * @brief Adds a new parameter. Accepts Type::IOPC_TYPECLASS_STRING, Type::IOPC_TYPECLASS_SIMPLE as weel as
	 * Type::IOPC_TYPECLASS_ENHANCED data types.
	 *
	 * If the @a varType is of @a Type::IOPC_TYPECLASS_ENHANCED, the @a EnhancedTypeParameter is created.
	 * @a SimpleTypeParameter is created for @a Type::IOPC_TYPECLASS_SIMPLE and @a Type::IOPC_TYPECLASS_ENHANCED
	 * type classes.
	 *
	 * If string types (Type::IOPC_TYPECLASS_STRING, EString or EWString), the [db.type.length] metadata must be specified either
	 * in the @a varType metadata or provided in the @a attributes input parameter.
	 *
	 * @throws IopcIncorrectUsageException if other type than @a Type::IOPC_TYPECLASS_ENHANCED, @a Type::IOPC_TYPECLASS_SIMPLE
	 * or @a Type::IOPC_TYPECLASS_ENHANCED  is specified.
	 *
	 * @param direction Parameter direction
	 * @param pos Position of the parameter in the prepared statement or in the result set, starting from 1.
	 * 	Note that some drivers (like the IopcOracle10g driver) require the parameters to be bound as a sequence so that
	 *  no positions are skipped. If a parameter already exists on the @a pos position, it is overwritten.
	 * @param varAddr Pointer to the value being bound. The pointer must remain valid as long as the cursor is used.
	 * 	If Type::IOPC_TYPECLASS_STRING type is used, this pointer must point to a buffer containing
	 *  standard, zero-terminated C string representation [IN] or to a buffer into which a standard C string
	 *  will be fetched [OUT]. So even if the varType is TypeDesc<std::string>
	 *  the varAddr must point [IN] or will point [OUT] to a standard C string.
	 *  If EString or EWString is used, the pointer should point to their instances.
	 * @param varType Parameter data type
	 * @param attributes Additional attributes or driver-dependent data.
	 * <ul>
	 * <li>[db.type.length] - For the @a EString and @a EWString types it specifies the
	 *     (maximum) length of the externalized string buffer - see the externalize methods in @a EnhancedTypeBase.
	 *     For the Type::IOPC_TYPECLASS_STRING types it specifies the
	 *     maximum allocated length (in characters without the trailing zero) for the input and output buffers.</li>
	 * </ul>
	 */
	void addParam(Parameter::ParamType direction, const int pos, const void* varAddr, const Type& varType, const MetadataHolder& attributes = MetadataHolder::emptyMeta);

	/**
	 * @brief Adds a new Enhanced data type parameter
	 *
	 * @param direction Parameter direction
	 * @param pos Position of the parameter in the prepared statement, starting from 1.
	 * 	If a parameter already exists on the @a pos position, it is overwritten.
	 * @param obj Reference to the value being bound. The reference must be valid as long as the cursor is used.
	 * @param attributes Additional attributes or driver-dependent data.
	 * <ul>
	 * <li>[db.type.length] - For the @a EString and @a EWString types it specifies the
	 *     (maximum) length of the externalized string buffer - see the externalize methods in @a EnhancedTypeBase.</li>
	 * </ul>
	 *
	 * If @a EString or @a EWString type is used, the [db.type.length] metadata must be specified either
	 * in the TypeDesc<EString>::getType() or in the TypeDesc<EWString>::getType metadata or
	 * provided in the @a attributes input parameter.
	 */
	void addEnhancedParam(Parameter::ParamType direction, const int pos,  EnhancedTypeBase& obj, const MetadataHolder& attributes = MetadataHolder::emptyMeta);

	Cursor(CursorImpl* cursorImpl, const std::string& sql);
	~Cursor();

	/**
	 * @brief Returns true if the @a Cursor state is not @a Cursor::CLOSED
	 */
	inline bool isOpened() { return state != CLOSED; }

	/**
	 * @brief Returns the current state of the @a Cursor
	 *
	 * @see State
	 */
	inline State getState() { return state; }

	/**
	 * @brief Returns the driver-side part of the @a Cursor implemenation.
	 */
	inline CursorImpl* getImplementation() { return implementation; }

	/**
	 * @brief Opens the cursor
	 *
	 * Database driver allocates all resources needed for the cursor execution
	 *
	 * Changes the cursor state to @a Cursor::OPENED
	 *
	 * @throws IopcDbException If the cursor is not in the @a Curosr::CLOSED state
	 */
	void open();

	/**
	 * @brief Prepares the SQL statement for execution
	 *
	 * Changes the cursor state to @a Cursor::PREPARED
	 *
	 * If the cursor satate is @a Cursor::CLOSED, it is opened by calling @a open().
	 */
	void prepare();

	/**
	 * @brief Binds a database NULL parameter to a place in the prepared statement
	 *
	 * @a Prepare() must be invoked before calling this method
	 *
	 * @param pos Position of the parameter in the prepared statement
	 * @param varType Data type of the parameter. @a varType must refer only to basic datatypes (IOPC_TYPECLASS_SIMPLE)
	 *   or to string datatypes (IOPC_TYPECLASS_STRING).
	 * @param attributes Additional driver dependent parameters.
	 *
	 * @throws IopcDbException If the cursor is not in @a Cursor::PREPARED state
	 */
	void preExecBindNull(const int pos, const Type& varType, const MetadataHolder& attributes = MetadataHolder::emptyMeta);

	/**
	 * @brief Binds a parameter to a place in the prepared statement
	 *
	 * @a Prepare() must be invoked before calling this method
	 *
	 * @param pos Position of the parameter in the prepared statement
	 * @param varAddr Address of the parameter value to be bound
	 * @param size Memory size of the parameter in bytes
	 * @param varType Data type of the parameter. @a varType must refer only to basic datatypes (IOPC_TYPECLASS_SIMPLE)
	 *   or to string datatypes (IOPC_TYPECLASS_STRING). If the IOPC_TYPECLASS_STRING type is specified, the
	 *   varAddr refers to standard zero-terminaded C string representation.
	 * @param attributes Additional driver dependent parameters.
	 *
	 * @throws IopcDbException If the cursor is not in @a Cursor::PREPARED state
	 */
	void preExecBind(const int pos, const void* varAddr, const unsigned int size, const Type& varType, const MetadataHolder& attributes = MetadataHolder::emptyMeta);

	/**
	 * @brief Executes the SQL statement
	 *
	 *  Performs the following steps:
	 *  <ol>
	 *   <li>If the cursor state is @a Cursor::CLOSED, the cursor is opened and prepared.
	 *	If the cursor state is @a Cursor::OPENED or @a Cursor::EXECUTED, the cursor is prepared.</li>
	 *	 <li>@a Parameter::preExec() is invoked on all IN parameters</li>
	 *	 <li>The SQL statement is executed</li>
	 *	 <li>Cursor state is changed to @a Cursor::EXECUTED
	 *	 <li>@a Parameter::postExec() is invoked on all IN parameters</li>
	 *	</ol>
	 */
	void execute();

	/**
	 * @brief Binds an output parameter to a place in the result set
	 *
	 * @param pos Position of the parameter in the result set
	 * @param varAddr Address of the output buffer where the fetched value will be stored.
	 * @param size Buffer size allocated to hold the fetched value in bytes
	 * @param varType Data type of the parameter. @a varType refers only to basic datatypes (IOPC_TYPECLASS_SIMPLE)
	 *   or to string datatypes (IOPC_TYPECLASS_STRING).
	 * @param isNull Pointer to an indicator variable which will be set to true if the fetched value is database NULL.
	 * 	 A NULL pointer may be passed to ignore the database NULL value notifiaction.
	 * @param attributes Additional driver dependent parameters.
	 */
	void preFetchBind(const int pos, const void* varAddr, const unsigned int size, const Type& varType, bool* isNull = NULL, const MetadataHolder& attributes = MetadataHolder::emptyMeta);

	/**
	 * @brief Fetches next result from the result set.
	 *
	 * Performs the following steps:
	 * <ol>
	 *  <li>Before first fetch after @a Cursor::execute() is performed, @a Parameter::preFetch()
	 *    is invoked on all OUT parameters.</li>
	 *  <li>Next result is fetched, output parameters and NULL indicators are updated.</li>
	 *  <li>@a Parameter::postFetch() is invoked on all OUT parameters.</li>
     * </ol>
     *
     * @returns true if a row was fetched. Otherwise, the OUT parameters are not updated and the method returns false.
	 * @throws IopcDbException if the cursor is not in the @a Cursor::EXECUTED state
	 */
	bool fetchNext();

	/**
	 * @brief Fetches previous result from the result set.
	 *
	 * At least one call to @a fetchNext() must be performed before calling @a fetchPrev.
	 * Output parameters and NULL indicators are updated and @a Parameter::postFetch() is invoked on all OUT parameters.
	 *
	 * @throws IopcDbException if the cursor is not in the @a Cursor::EXECUTED state
	 * @throws IopcNotSupportedException if this operation is not supported by the database driver.
	 */
	bool fetchPrev();

	/**
	 * @brief Returns current position in the result set
	 */
	long getPosition();

	/**
	 * @brief Closes the cursor
	 *
	 * Database driver release all resources allocated in the @a Cursor::open() method
	 * Changes the @a Cursor state to @a Cursor::CLOSED
	 */
	void close();

//	virtual void test() = 0;


};

} // namespace iopc

#endif /*CURSOR_H_*/




