/**
 * @file driverExtension.h
 *
 * @author Petr Cermak
 *
 * @ingroup iopcdb
 */

#ifndef DRIVEREXTENSION_H_
#define DRIVEREXTENSION_H_

#include "iopccommon/defs.h"
#include <string>

namespace iopc {

class Driver;

/**
 * @brief Base class of all driver extensions
 *
 * Driver extensions provide additional functionalities to drivers.
 *
 * @ingroup iopcdb
 */
class DLL_EXTERNAL DriverExtension
{
	friend class Driver;
protected:

	/**
	 * @brief Registers the driver extension and associates it with a specified driver
	 *
	 * @param driverName The driver associated with this extension
	 * @param name Name of this extension. The name must be unique across all extensions of the same driver.
	 */
	DriverExtension(const std::string& driverName, const std::string& name);
	virtual ~DriverExtension();

	/**
	 * @brief Extension initialization routine
	 *
	 * Usually it registers additional features to the driver by calling:
	 * @code
	 * driver.registerFeature<Feature>(featureImpl);
	 * @endcode
	 *
	 * @param driver The driver associated with this feature
	 */
	virtual void doInit(Driver& driver) =0;

	/**
	 * @brief Returns name of the associated driver
	 */
	inline std::string& getDriverName() { return driverName; }

	/**
	 * @brief Returns name of this feature
	 */
	inline std::string& getName() { return name; }
private:
	std::string driverName;
	std::string name;
};


}

#endif /*DRIVEREXTENSION_H_*/
