/**
 * @file cache.h
 *
 * @author Jan Hadrava
 * @date 2004
 * @ingroup iopclibcache
 * @defgroup iopclibcache iopclib-cache
 */

#ifndef CACHE_H_
#define CACHE_H_

#include "cacheLock.h"
//xxx#include "tDbPtr.h"
#include "../persIdentification.h"
#include "iopccommon/defsPlatform.h"
#include "iopcdb/connection.h"
#include "iopccommon/exceptionBase.h"
#include "../databaseObject.h"

namespace iopc {

/**
 * @brief Object cache interface.
 *
 * @ingroup iopclibcache
 */
class DLL_EXTERNAL Cache {
//friend class Connection;
public:
	inline Cache();
	inline virtual ~Cache();

	/**
	 * @brief Sets the connection associated with cache.
	 *
	 * @param conn The connection.
	 */
	inline virtual void setConnection(Connection* conn = NULL);

	inline virtual Connection& getConnection();

	inline bool isConnectionSet() { return conn != NULL; };


	/**
	 * @brief Synchronizes states of the cache and database.
	 *
	 * Removes all objects. Makes it possible to harmlessly break
	 * association of the cache and its connection.
	 */
	virtual void clear() = 0;

	/**
	 * @brief Saves an object into database.
	 *
	 * Locks it in cache.
	 *
	 * @param persId ID of the object.
	 * @param obj Object to be made persistent.
	 *
	 * @return An instance representing locked object and making it possible to read and modify the object.
	 */
	virtual CacheLock& makePersistent(const PersIdentification& persId, DatabaseObject& objectPtr) = 0;

	/**
	 *
	 * @brief Assures an object is in cache, locks it and returns an instance making it possible to
	 * access the object.
	 *
	 * @param persId ID of the object.
	 * @param rdonly Indicates the type of required access to the object (read-only/write).
	 *
	 * @return An instance making it possible to access the object.
	 */
	virtual CacheLock& getCacheLock(const PersIdentification& persId, bool rdonly = false) = 0;

	/**
	 * @brief Deletes an object.
	 *
	 * @param persId ID of the object.
	 */
	virtual void dbDelete(const PersIdentification& persId) = 0;

	/**
	 * @brief Represents a hint for the cache to remove local copy of an object.
	 *
	 * Can be ignored.
	 *
	 * @param persId ID of the object.
	 */
	virtual void cacheDelete(const PersIdentification& persId) = 0;

	/**
	 * @brief Synchronize the state of cache and DB by updating the DB.
	 */
	virtual void updateDirty() = 0;

	/**
	 * @brief Synchronize the state of cache and DB by discarding all kept object
	 * (without modifications to DB).
	 */
	virtual void removeAll() = 0;


	/**
	 * @brief Creates a copy of the current cache
	 */
	virtual Cache* clone() = 0;


protected:
	Connection *conn;	/**< A connection associated with the cache. */
};

inline Cache::Cache()
: conn(NULL)
{
}

inline Cache::~Cache()
{
	if (conn != NULL) {
		IOPC_LOG_THROW(IopcIncorrectUsageException("Destroying cache that is still in use"));
	}
}

inline void Cache::setConnection(class Connection *conn)
{
	if (this->conn == NULL) {
		this->conn = conn;
	} else if (conn == NULL) {
		clear();
		this->conn = NULL;
	} else if (this->conn != conn) {
		IOPC_LOG_THROW(IopcIncorrectUsageException("A connection is trying to claim cache that's already in use"));
	}
}

inline Connection& Cache::getConnection()
{
	if (this->conn == NULL) {
		IOPC_LOG_THROW(IopcIncorrectUsageException("No connection is set. Call setConnection first."));
	}
	return *conn;
}

} /*namespace iopc*/

#endif /*CACHE_H_*/
