/**
 * @file extendedCache.h
 *
 * @author Jan Hadrava
 * @date 2004
 * @ingroup iopclibcache
 */

#ifndef EXTENDEDCACHE_H_
#define EXTENDEDCACHE_H_

// project: POLiTe
// authors: Jan Hadrava 2004

#include <map>
#ifndef WIN32
#include <ext/hash_map>
#endif /*WIN32*/

#include "iopccommon/defs.h"
#include "cache.h"
#include "cacheLock.h"
#include "strategy.h"
#include "strategySelector.h"
#include "iopccommon/timeStamp.h"
#include "../persIdentification.h"

namespace iopc {

/** Common predecessor of so-called extended caches. Provides interface which makes it possible
 * to specify db access strategies when using some methods. Includes interface for external maintenance
 * of the cache.
 *
 * @ingroup iopclibcache
 */
class DLL_EXTERNAL ExtendedCache : public Cache {
	friend class ComposedCache;
public:
	/** Initialize cache with the same strategies for all object.
	* \param minValue Minimal value of objects obliged to stay in the cache.
	* \param maxCost Maximal total cost of stored objects.
	* \param strat Default strategy.
	*/
	inline ExtendedCache(long minValue = MINLONG, long maxCost = MAXLONG, const Strategy& strat = Strategy::defaultStrategy);
	/** Initialize cache with strategies passed by selector.
	* \param minValue Minimal value of objects obliged to stay in the cache.
	* \param maxCost Maximal total cost of stored objects.
	* \param strat strategy selector.
	**/
	inline ExtendedCache(long minValue, long maxCost, StrategySelector* stratSel);

	inline virtual ~ExtendedCache();

private:
	ExtendedCache(ExtendedCache& );
	ExtendedCache& operator=(ExtendedCache& );

protected:
	inline ExtendedCache(long minValue, long maxCost, StrategySelector* stratSel, Strategy strat);

	inline virtual void clear();

	inline virtual CacheLock& makePersistent(const PersIdentification& persId, DatabaseObject& obj);
	inline virtual CacheLock& getCacheLock(const PersIdentification& persId, bool rdonly = false);

	inline virtual void dbDelete(const PersIdentification& persId);
	inline virtual void cacheDelete(const PersIdentification& persId);

	inline virtual void updateDirty();
	inline virtual void removeAll();

protected:
	/** Synchronizes states of the cache and database. Removes all objects. Makes it possible to
	* harmlessly break association of the cache and its connection.
	* Uses strategy selector passed by caller.
	* \param strSel The strategy selector.
	*/
	virtual void clear(const StrategySelector& strSel) = 0;
	/** Saves an object into database. Locks it in cache.
	* Uses strategy selector passed by caller.
	* \param strSel The strategy selector.
	* \param persId ID of the object.
	* \param obj Object to be made persistent.
	* \return An instance representing locked object and making it possible to read and modify the object.
	*/
	virtual CacheLock& makePersistent(const StrategySelector& strSel, const PersIdentification& persId, DatabaseObject& objectPtr) = 0;
	/** Assures an object is in cache, locks it and returns an instance making it possible to
	* access the object.
	* Uses strategy selector passed by caller.
	* \param strSel The strategy selector.
	* \param persId ID of the object.
	* \param rdonly Indicates the type of required access to the object (read-only/write).
	* \return An instance making it possible to access the object.
	*/
	virtual CacheLock& getCacheLock(const StrategySelector& strSel, const PersIdentification& persId, bool rdonly = false) = 0;
	/** Deletes an object.
	* Uses strategy selector passed by caller.
	* \param strSel The strategy selector.
	* \param persId ID of the object.
	*/
	virtual void dbDelete(const StrategySelector& strSel, const PersIdentification& persId) = 0;
	/** Represents a hint for the cache to remove local copy of an object. Can be ignored.
	* Uses strategy selector passed by caller.
	* \param strSel The strategy selector.
	* \param persId ID of the object.
	*/
	virtual void cacheDelete(const StrategySelector& strSel, const PersIdentification& persId) = 0;
	/** Synchronize the state of cache and DB by updating the DB.
	* Uses strategy selector passed by caller.
	* \param strSel The strategy selector.
	*/
	virtual void updateDirty(const StrategySelector& strSel) = 0;
	/** Synchronize the state of cache and DB by discarding all kept object
	(without modifications to DB).http://www-ucjf.troja.mff.cuni.cz/scheirich/comics/mff_life_09.jpg
	* Uses strategy selector passed by caller.
	* \param strSel The strategy selector.
	*/
	virtual void removeAll(const StrategySelector& strSel) = 0;

public:
	/** Returns total cost of objects which can be removed from cache (i.e. are not locked).
	*/
	inline virtual long getActiveCost();
	/** Returns total cost of objects which can't be removed from cache (i.e. are locked).
	*/
	inline virtual long getLockedCost();
	/** Returns value of the object having minimal value.
	*/
	inline virtual long getWorstValue(const TimeStamp& current);
	/** Removes the object with minimal value.
	*/
	inline virtual long removeWorst();

	/** Function executing one elementar maintenance step
	** \return false when there's no work for maintainer, true otherwise.
	*/
	inline virtual bool maintainerStep(long& costChange);

	/** Gets the time it's possible to sleep for the external maintainer (unless it's woken up).
	** \return Maximal number of milliseconds the cache doesn't need to be externally maintained for.
	*/
	inline virtual long getSleepTime(const TimeStamp& current, long minValue);

	/** Function imposing specified limits on the cache. The limits are fullfilled just
	* for the time of the call.
	* \param minValue Minimal value of an element being in the cache.
	* \param maxCost Maximal total cost of elements in the cache.
	* \return false if specified limits cannot be achieved (e.g. due to locks), true otherwise (on success}.
	*/
	virtual bool imposeLimits(long minValue, long maxCost);
	/** Imposes default limits (specified by attributes). */
	inline virtual bool imposeLimits();

	/** Sets limits permanently (until next explicit change).
	* \param minValue MiStrategynimal value of an element being in the cache
	* \param maxCost Maximal total cost of elements in the cache.
	* \return false if specified limits cannot be achived for time of the call, true otherwise (on success);
	** even if false is returned the limits are set as specified and should be achieved in future (when possible).
	*/
	inline virtual bool setLimits(long minValue, long maxCost);
protected:
	Strategy defaultStrategy; /**< Default strategy */
	StrategySelector defaultStratSel; /**< Default strategy selector. */
	StrategySelector* stratSel; /**< Strategy selector created on cache initialization. */
	long minValue; /**< Minimal value of object obliged to stay in cache. */
	long maxCost; /**< Maximal total cost of objects being in cache. */
};

inline ExtendedCache::ExtendedCache(long minValue, long maxCost, StrategySelector* stratSel)
: defaultStratSel(), stratSel(stratSel), minValue(minValue), maxCost(maxCost)
{
}

inline ExtendedCache::ExtendedCache(long minValue, long maxCost, const Strategy& strat)
: defaultStrategy(strat), defaultStratSel(strat), stratSel(NULL), minValue(minValue), maxCost(maxCost)
{
}

inline ExtendedCache::ExtendedCache(long minValue, long maxCost, StrategySelector* stratSel, Strategy strat)
: defaultStrategy(strat), defaultStratSel(strat), stratSel(stratSel), minValue(minValue), maxCost(maxCost)
{
}

inline ExtendedCache::~ExtendedCache()
{
}

inline void ExtendedCache::clear()
{
	clear(stratSel != NULL ? *stratSel : defaultStratSel);
}

inline CacheLock& ExtendedCache::makePersistent(const PersIdentification& persId, DatabaseObject& objectPtr)
{
	return makePersistent(stratSel != NULL ? *stratSel : defaultStratSel, persId, objectPtr);
}

inline CacheLock& ExtendedCache::getCacheLock(const PersIdentification& persId, bool rdonly)
{
	return getCacheLock(stratSel != NULL ? *stratSel : defaultStratSel, persId, rdonly);
}

inline void ExtendedCache::dbDelete(const PersIdentification& persId)
{
	dbDelete(stratSel != NULL ? *stratSel : defaultStratSel, persId);
}

inline void ExtendedCache::cacheDelete(const PersIdentification& persId)
{
	cacheDelete(stratSel != NULL ? *stratSel : defaultStratSel, persId);
}

inline void ExtendedCache::updateDirty()
{
	updateDirty(stratSel != NULL ? *stratSel : defaultStratSel);
}

inline void ExtendedCache::removeAll()
{
	removeAll(stratSel != NULL ? *stratSel : defaultStratSel);
}

inline long ExtendedCache::getActiveCost()
{
	return 0;
}

inline long ExtendedCache::getLockedCost()
{
	return 0;
}

inline long ExtendedCache::getWorstValue(const TimeStamp& current)
{
	return MAXLONG;		/* best value possible */
}

inline long ExtendedCache::removeWorst()
{
	return 0;		/* unable to remove item */
}

inline bool ExtendedCache::maintainerStep(long& costChange)
{
	return false;		/* unable to remove anything */
}

inline long ExtendedCache::getSleepTime(const TimeStamp& current, long minValue)
{
	return 1800*1000000;	/* sleep forever */
}

inline bool ExtendedCache::imposeLimits()
{
	return imposeLimits(minValue, maxCost);
}

inline bool ExtendedCache::setLimits(long minValue, long maxCost)
{
	this->minValue = minValue;
	this->maxCost = maxCost;
	return imposeLimits(minValue, maxCost);
}

}/*namespace iopc*/

#endif /*EXTENDEDCACHE_H_*/
