/**
 * @file voidCache.h
 *
 * @author Jan Hadrava
 * @date 2004
 * @ingroup iopclibcache
 */

#ifndef VOIDCACHE_H_
#define VOIDCACHE_H_

#include <map>
#include "cache.h"
#include "iopccommon/defs.h"
#include "iopccommon/rwLock.h"
#include "iopccommon/condVar.h"

namespace iopc {

class VoidCacheItem;

/**
 * @brief Simple cache implementation removing objects as soon as possible (immediately after unlocking last lock).
 *
 * Supports multithreaded usage. Objects are represented by VoidCacheItem. Instances of that
 * class are stored in a map container provided by STL.
 *
 * @ingroup iopclibcache
 */
class DLL_EXTERNAL VoidCache : public Cache {

public:
	VoidCache();
	~VoidCache();

private:
	typedef std::map<PersIdentification, VoidCacheItem*> ContType;
	typedef ContType::iterator IterType;

	VoidCache(const VoidCache&);
	//VoidCache &operator=(const VoidCache &);

public:
	/* used by Connection */
	/* TODO: vyjimka... muze.... called in connection destructor -> can't throw exception */
	virtual void clear();

	virtual CacheLock& makePersistent(const PersIdentification& persId, DatabaseObject& objectPtr);
	virtual CacheLock& getCacheLock(const PersIdentification& persId, bool rdonly = false);

	virtual void dbDelete(const PersIdentification& persId);
	virtual void cacheDelete(const PersIdentification& persId);

	virtual void updateDirty();
	virtual void removeAll();

	/* used by VoidCacheItem */
	/** Called when unlocking object lock, tries to remove the object from cache
	(in case that there are no locks anymore).
	If succesfully inserted to the database, persId becomes valid and receives generated OID or
	key values
	*/
	void tryCacheDelete(PersIdentification& persId);

	virtual Cache* clone();

private:
	long items; /**< Number of objects present in cache. */
	ContType cont; /**< Container holding locked items. */
	RWLock contLock; /**< Locks container when locking/unlocking its items. */
	CondVar emptyCond;
};

}


#endif /*VOIDCACHE_H_*/
