/**
 * @file cachePtr.h
 *
 * @author Jan Hadrava
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */

#ifndef CACHEPTR_H_
#define CACHEPTR_H_

#include "cache/cacheLock.h"
#include "persIdentification.h"
#include "db/cachedConnection.h"
#include "iopccommon/countedPtr.h"

namespace iopc {

/**
 * @brief A cache pointer class
 *
 * The existence of the cache pointer guarantees
 * that the object is loaded into the memory and that the memory address of the object will not change until
 * the instance of the cache pointer is destroyed (unlocked). Such process locks the objects in the cache
 * so they can't be removed unless they are unlocked.
 *
 * @ingroup iopclib
 */
template<typename T> class CachePtr{

public:
	/**
	 * @brief Creates an invalid cache pointer
	 */
	CachePtr();

	/**
	 * @brief Crates a cache pointer to a persistent object
	 *
	 * Requested object is looked up in the cache associated with the @a conn connection
	 * and if not found the object is loaded from the underlying database.
	 *
	 * @param conn A connection with an associated cache in which the object should be locked
	 * @param persId Identity of the object to be locked
	 */
	CachePtr(CachedConnection& conn, const PersIdentification& persId);

	/**
	 * @brief Creates a cache pointer to a transient object owned by the @a DbPtr<T>
	 */
	CachePtr(CountedPtr<T> objectPtr);

	/**
	 * @brief Crates a cache pointer from an already locked object
	 */
	CachePtr(CacheLock& cLock);

	/**
	 * @brief Copy constructor
	 */
	CachePtr(const CachePtr<T>& cptr);

	~CachePtr();

	/**
	 * @brief Copy operator
	 */
	CachePtr<T> &operator=(const CachePtr<T>& src);

	/**
	 * @brief Releases the lock and invalidates the cache pointer
	 */
	void release();

	/**
	 * @brief Determines whether the pointer is valid or not. (Points to a cache-locked instance)
	 */
	inline bool isValid() const;

	/**
	 * @brief Determines whether the pointer is valid or not. (Points to a cache-locked instance)
	 */
	inline operator bool() const;

	/**
	 * @brief Dereferences cache pointer
	 *
	 * @return Counted pointer to the referenced object (points to NULL for invalid cache pointer).
	 */
	inline CountedPtr<T> operator*() const;

	/**
	 * @brief Dereferences cache pointer
	 *
	 * @return Counted pointer to the referenced object (points to NULL for invalid cache pointer).
	 */
	inline CountedPtr<T> operator->() const;

	/**
	 * @brief Dereferences cache pointer
	 *
	 * @return Counted pointer to the referenced object (points to NULL for invalid cache pointer).
	 */
	inline CountedPtr<T> getObjectPtr() const;

protected:
	/**
	 * @brief Counted pointer, for both persistent and transient instances.
	 */
	CountedPtr<T> objectPtr;


	/**
	 * @brief Cache item manipulator (NULL for transient instances).
	 */
	CacheLock* cLock;
};

template<typename T> CachePtr<T>::CachePtr()
{
	cLock = NULL;
	objectPtr = NULL;
	//countedObjectPtr = NULL;
}

template<typename T> CachePtr<T>::CachePtr(CachedConnection& conn, const PersIdentification& persId)
{
	cLock = &conn.getCacheLock(persId, false);
	objectPtr = CountedPtr<T>(dynamic_cast<T*>(&cLock->getObject()), false);
	//countedObjectPtr = NULL;
}

template<typename T> CachePtr<T>::CachePtr(CountedPtr<T> object)
{
	cLock = NULL;
	//this->objectPtr = NULL;
	this->objectPtr =  object;
	//dynamic_cast<T*>(objectPtr);
}

template<typename T> CachePtr<T>::CachePtr(const CachePtr<T>& src)
{
	objectPtr = src.objectPtr;
	//countedObjectPtr = src.countedObjectPtr;
	if (src.cLock == NULL) {	/* transient instance */
		cLock = NULL;
	} else {			/* persistent instance */
		cLock = &src.cLock->addLock(false);
	}
}

template<typename T> CachePtr<T>::CachePtr(CacheLock& cLock)
{
	this->cLock = &cLock;
	objectPtr = CountedPtr<T>(dynamic_cast<T*>(&cLock.getObject()), false);
	//countedObjectPtr = NULL;
}

template<typename T> CachePtr<T>::~CachePtr()
{
	IOPC_FUNC_INVOKED;
	release();
}

template<typename T> CachePtr<T>& CachePtr<T>::operator=(const CachePtr<T>& src)
{
	if (this == &src)
		return *this;
	objectPtr = src.objectPtr;
	//countedObjectPtr = src.countedObjectPtr;
	if (cLock != NULL) {
		cLock->rmLock();
	}
	if (src.cLock == NULL) {	/* transient instance */
		cLock = NULL;
	} else {			/* persistent instance */
		cLock = &src.cLock->addLock(false);
		IOPC_ASSERT(objectPtr == (T *)&cLock->getObject());
	}
	return *this;
}

template<typename T> void CachePtr<T>::release()
{
	IOPC_FUNC_INVOKED;
	if (cLock != NULL) {
		cLock->rmLock();
		cLock = NULL;
	}
	objectPtr = CountedPtr<T>();
	//countedObjectPtr = NULL;
	IOPC_FUNC_ENDED;
}

template<typename T> inline bool CachePtr<T>::isValid() const
{
	return objectPtr != NULL;// || countedObjectPtr != NULL;
}

template<typename T> inline CachePtr<T>::operator bool() const
{
	return isValid();
}

template<typename T> CountedPtr<T> CachePtr<T>::operator*() const
{
	return getObjectPtr();
}

template<typename T> CountedPtr<T> CachePtr<T>::operator->() const
{
	return getObjectPtr();
}

template<typename T> CountedPtr<T> CachePtr<T>::getObjectPtr() const
{
	return objectPtr;
}

} /*namespace*/



#endif /*CACHEPTR_H_*/
