/**
 * @file objectStatementsFeature.h
 *
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */

#ifndef OBJECTSTATEMENTSFEATURE_H_
#define OBJECTSTATEMENTSFEATURE_H_

#include "iopcdb/dbFeature.h"
#include "iopccommon/metadataHolder.h"
#include "../types/dbType.h"
#include "iopccommon/rwLock.h"
#include "../types/objectMappingStrategy.h"
#include <vector>

namespace iopc {

class Type;

/**
 * @brief This feature declares methods used by the ADT (object) mapping type
 *
 * DbType::getKeyAttributes() for object-mapped classes contain always only a single OID attribute.
 *
 * @ingroup iopclib
 */
class ObjectStatementsFeature : public DbFeature
{
private:
	static std::string featureName;
public:
	virtual const std::string& getNameInst() {
		return getName();
	}

	static const std::string& getName() {
		return featureName;
	}

	// interface
	/**
	 * @brief Returns a statement that selects from a Simple view associated with the type the @a currentStrategy represents
	 *
	 * Selected columns are specified by DbType::getAllPersistentAttributes() (currentStrategy.getDbType().getAllPersistentAttributes())
	 * The order of the columns selected must correspond to the order of attributes in the list.
	 * WHERE part of the select statement must generated from the DbType::getKeyAttributes() list or just by
	 * filtering by OID.
	 */
	virtual std::string getSelectFromClassViewStatement(ObjectMappingStrategy& currentStrategy, const MetadataHolder& attributes = MetadataHolder::emptyMeta, bool useCache = true) =0;

	/**
	 * @brief Returns a statement that selects from a database table or ADT type associated with the type the @a currentStrategy represents
	 *
	 * Selected columns are specified by ObjectMappingStrategy::getMappedAttributes() (currentStrategy.getMappedAttributes())
	 * The order of the columns selected must correspond to the order of attributes in the list.
	 * WHERE part of the select statement must generated from the DbType::getKeyAttributes() list or just by
	 * filtering by OID.
	 */
	virtual std::string getSelectFromClassTableStatement(ObjectMappingStrategy& currentStrategy, const MetadataHolder& attributes = MetadataHolder::emptyMeta, bool useCache = true) =0;

	/**
	 * @brief Returns a statement that inserts data to a database table using ADT associated with the type the @a currentStrategy represents
	 *
	 * The insert must define input parameters that will correspond with attributes stored in the ObjectMappingStrategy::getMappedAttributes() list.
	 */
	virtual std::string getInsertStatement(ObjectMappingStrategy& currentStrategy, const MetadataHolder& attributes = MetadataHolder::emptyMeta, bool useCache = true) =0;

	/**
	 * @brief Returns a statement that updates data in a database table associated with the type the @a currentStrategy represents
	 *
	 * The update must define input parameters that will correspond with attributes stored in the ObjectMappingStrategy::getMappedAttributes() list.
	 * Followed by parameters that define which rows to update defined by DbType::getKeyAttributes() list or just by
	 * an OID column.
	 */
	virtual std::string getUpdateStatement(ObjectMappingStrategy& currentStrategy, const MetadataHolder& attributes = MetadataHolder::emptyMeta, bool useCache = true) =0;

	/**
	 * @brief Returns a statement that deletes data from a database table associated with the type the @a currentStrategy represents
	 *
	 * The statement must define input parameters that restrict the delete operation. (OID or DbType::getKeyAttributes())
	 */
	virtual std::string getDeleteStatement(ObjectMappingStrategy& currentStrategy, const MetadataHolder& attributes = MetadataHolder::emptyMeta, bool useCache = true) =0;

	/**
	 * @brief Performs a driver-side validation of the persistent object model and metadata
	 *
	 * @param[in] currentStrategy A mapping strategy object representing a type being mapped.
	 * @param[out] msg Error messages generated by the database driver.
	 *
	 * @returns true - validation completed successfully, false otherwiese. If false is returend, msg contains relevant error message.
	 */
	virtual bool validate(ObjectMappingStrategy& currentStrategy, std::string& msg) =0;
};


}




#endif /*OBJECTSTATEMENTSFEATURE_H_*/
