/**
 * @file orStatementsFeature.h
 *
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */
#ifndef ORSTATEMENTSFEATURE_H_
#define ORSTATEMENTSFEATURE_H_

#include "iopcdb/dbFeature.h"
#include "iopccommon/metadataHolder.h"
#include "../types/dbType.h"
#include "iopccommon/rwLock.h"
#include "../types/orMappingStrategy.h"
#include <vector>

namespace iopc {

class Type;

/**
 * @brief This feature declares methods used by the O/R (vertical,  horizontal, filtered) mapping type
 *
 * @ingroup iopclib
 */
class ORStatementsFeature : public DbFeature
{
private:
	static std::string featureName;
public:
	virtual const std::string& getNameInst() {
		return getName();
	}

	static const std::string& getName() {
		return featureName;
	}

	// interface
	/**
	 * @brief Returns a statement that selects from a Simple view associated with the type the @a currentStrategy represents
	 *
	 * Selected columns are specified by DbType::getAllPersistentAttributes() (currentStrategy.getDbType().getAllPersistentAttributes())
	 * The order of the columns selected must correspond to the order of attributes in the list.
	 * WHERE part of the select statement must be restricted by input parameters generated from the DbType::getKeyAttributes() list.
	 * (currentStrag.getDbType().getKeyAttributes()). Again in the corresponding order.
	 */
	virtual std::string getSelectFromClassViewStatement(ORMappingStrategy& currentStrategy, const MetadataHolder& attributes = MetadataHolder::emptyMeta, bool useCache = true) =0;

	/**
	 * @brief Returns a statement that selects from a database table associated with the type the @a currentStrategy represents
	 *
	 * Selected columns are specified by DbType::getMappedAttributes() (currentStrategy.getDbType.getMappedAttributes())
	 * The order of the columns selected must correspond to the order of attributes in the list.
	 * WHERE part of the select statement must be restricted by input parameters generated from the DbType::getKeyAttributes() list.
	 * (currentStrag.getDbType().getKeyAttributes()). Again in the corresponding order.
	 */
	virtual std::string getSelectFromClassTableStatement(ORMappingStrategy& currentStrategy, const MetadataHolder& attributes = MetadataHolder::emptyMeta, bool useCache = true) =0;

	/**
	 * @brief Returns a statement that insert data to a database table associated with the type the @a currentStrategy represents
	 *
	 * The insert must define input parameters that will correspond with attributes stored in the DbType::getMappedAttributes() list.
	 */
	virtual std::string getInsertStatement(ORMappingStrategy& currentStrategy, const MetadataHolder& attributes = MetadataHolder::emptyMeta, bool useCache = true) =0;

	/**
	 * @brief Returns a statement that updates data in a database table associated with the type the @a currentStrategy represents
	 *
	 * The update must define input parameters that will correspond with attributes stored in the DbType::getMappedAttributes() list.
	 * Followed by the WHERE clause generated from the DbType::getKeyAttributes() list.
	 */
	virtual std::string getUpdateStatement(ORMappingStrategy& currentStrategy, const MetadataHolder& attributes = MetadataHolder::emptyMeta, bool useCache = true) =0;

	/**
	 * @brief Returns a statement that updates data in a database table associated with the type the @a currentStrategy represents
	 *
	 * The update must define input parameters that will correspond with attributes stored in the @a updateCols list.
	 * Followed by the WHERE clause generated from the DbType::getKeyAttributes() list.
	 */
	virtual std::string getUpdateStatement(ORMappingStrategy& currentStrategy, const ORMappingStrategy::MappedAttributesList& updateCols, const MetadataHolder& attributes = MetadataHolder::emptyMeta, bool useCache = true) =0;

	/**
	 * @brief Returns a statement that deletes data from a database table associated with the type the @a currentStrategy represents
	 *
	 * WHERE part of the delete statement must be restricted by input parameters generated from the DbType::getKeyAttributes() list.
	 */
	virtual std::string getDeleteStatement(ORMappingStrategy& currentStrategy, const MetadataHolder& attributes = MetadataHolder::emptyMeta, bool useCache = true) =0;

	/**
	 * @brief Performs a driver-side validation of the persistent object model and metadata
	 *
	 * @param[in] currentStrategy A mapping strategy object representing a type being mapped.
	 * @param[out] msg Error messages generated by the database driver.
	 *
	 * @returns true - validation completed successfully, false otherwiese. If false is returend, msg contains relevant error message.
	 */
	virtual bool validate(ORMappingStrategy& currentStrategy, std::string& msg) =0;
};


}




#endif /*ORSTATEMENTSFEATURE_H_*/
