/**
 * @file persIdentification.h
 *
 * @author Jan Hadrava
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */

#ifndef PERSIDENTIFICATION_H_
#define PERSIDENTIFICATION_H_

#include <string>
#include "types/dbType.h"
#include "iopccommon/defs.h"
#include "oidObject.h"
#include "iopcdb/connection.h"
#include "keyValues.h"

#ifndef WIN32
//#include <unordered_map>
#include <ext/hash_map>
#endif /*WIN32*/


namespace iopc {

/**
 * @brief Encapsulates persistent object identity and provides access to O/R mapping services
 *
 * O/R mapping services are provided using @a DbType.
 *
 * @ingroup iopclib
 */
class DLL_EXTERNAL PersIdentification {
	friend class DbPtrBase;
public:
	PersIdentification();
	PersIdentification(const Type& type);
	PersIdentification(const PersIdentification& src);
	PersIdentification(const Type& type, const OidType& oid, bool loadRealType = true);
	PersIdentification(const Type& type, const KeyValues& keyValues);
	~PersIdentification();
	PersIdentification& operator=(const PersIdentification& src);
	bool operator<(const PersIdentification& id) const;	/* needed by STL containers */
	bool operator==(const PersIdentification& id) const;
	bool operator!=(const PersIdentification& id) const;
	int compare(const PersIdentification& id) const;

public:	/* methods to be used by cache implementations */

	/**
	 * @brief Loads object from a database.
	 *
	 * @param conn Connection to be used for DB operation.
	 * @param wait Specifies whether the operation should wiat for database locks to be unlocked.
	 * @param lock Specifies whether the persistent copy should be locked in DB exlusively when loaded.
	 * @return Loaded object.
	 */
	DatabaseObject* load(Connection& conn, bool wait, bool lock) const;


	/**
	 * @brief Loads object data from a cursor
	 *
	 * @param cur Cursof from which to load the data
	 *
	 * return Laoded object
	 */
	DatabaseObject* load(Cursor& cur) const;

	/**
	 * @brief Deletes object from a database.
	 *
	 * @param conn Connection used for the DB operation.
	 */
	void remove(Connection& conn) const;

	/**
	 * @brief Updates persistent copy
	 *
	 * @param conn Connection used for the DB operation.
	 * @param obj Local copy that is used to update the persistent copy in the database.
	 */
	void update(Connection& conn, DatabaseObject& obj) const;

	/**
	 * @brief Inserts object into a DB.
	 *
	 * @param conn Connection used for the DB operation.
	 * @param obj Persistent objects that is inserted into the database
	 */
	void insert(Connection& conn, DatabaseObject& obj) const;

	// virtualise
	//void setSuperType(Connection& conn);

	/**
	 * @brief Creates an identity for a transient object
	 *
	 * For OID object it selects a new OID from the OID sequence.
	 * For a non-OID persistent object it extracts KeyValues from the @a obj instance attributes.
	 *
	 * @param conn Connection used for a new OID retrieval
	 * @param obj The transient object for which an identity is created
	 */
	void initIdentity(Connection& conn, DatabaseObject& obj);

	/**
	 * @brief Compares serialId timestamp of an object with a timestamp
	 * stored in the oid-classname catalogue in database
	 *
	 * @param conn Connection used to retrieve the @a obj database timestamp
	 * @param obj Obj with the serialId timestamp to compare
	 */
	bool versionMatches(Connection& conn, OidObject& obj) const;



public:

	/**
	 * @brief Returns @a KeyValues for the current object.
	 *
	 * May be empty if @a PersIdentification represents an OID object identity or is not yet initialized
	 */
	inline const KeyValues& getKeyValues() const { return keyValues; }

	/**
	 * @brief Returns @a KeyValues for the current object.
	 *
	 * Undefined if @a PersIdentification represents a non-OID object identity or is not yet initialized
	 */
	inline const OidType getOid() const { return oid; }

	inline OidType& getOidRef() { return oid; }

	inline const OidType& getConstOidRef() const { return oid; }

	//inline VirtualOidType getVirtualOid() const { return virtualOid; }

	/**
	 * @brief Returns a type for which the current @a PersIdentification was created
	 *
	 * @returns TypeDesc<OidObject> if the identity has not been initialized with any type yet
	 */
	inline const Type& getType() const { return dbType.getType(); }

	/**
	 * @brief Returns a @a DbType instance associated with type for which the current @a PersIdentification was created
	 *
	 * @returns @a OidObject @a DbType if the identity has not been initialized with any type yet
	 */
	inline const DbType& getDbType() const { return dbType; }

	/**
	 * @brief Determines whether the current instance represents any object identity or not.
	 */
	inline bool isValid() const { return valid && usable; }


	/**
	 * @brief Type of the encapsulated identity
	 */
	enum IDType {
		OID = 0,
		KEYVALUES
	};

	/**
	 * @brief Returns the type of the encapsulated identity
	 */
	inline IDType getIDType() const { return idType; }

private:
	/**
	 * Returns real type if oid was specified and loadRealType was set true during initialization.
	 * Otherwise returns the current dbType.
	 */
	DbType& getDesiredDbType(Connection& conn) const;

	void invalidate();

	void setValid();

private:

	KeyValues keyValues;

	bool valid;

	bool usable;

	IDType idType;

	OidType oid;


	/**
	 * @brief The O/R mapping wrapper for a @a Type
	 */
	DbType& dbType;

	/**
	 * @brief Determines whether to search for actual type of an OID object (polymorphic load)
	 */
	bool loadRealType;

	/**
	 * @brief Cached real dbType, otherwise NULL
	 */
	mutable DbType* realDbType;

public:

	const static PersIdentification nil;
};

} /* iopc */

#ifndef WIN32
namespace __gnu_cxx {

/* this functor is needed by hash_table using PersIdentification as type of its keys. */
template<> struct hash<iopc::PersIdentification>
{
	size_t operator()(const iopc::PersIdentification& persId) const {
		size_t retval = 0;
		//std::string str = persId.getType().getQualifiedName();
		hash<const char*> H;
		hash<iopc::OidTypeInner> HOid;

		if (persId.getIDType() == iopc::PersIdentification::OID) {
			retval += H("oid");
			retval += HOid(persId.getOid());
		} else {
			for (iopc::KeyValues::Iterator it = persId.getKeyValues().begin(); it != persId.getKeyValues().end(); it++) {
				retval += H(it->first.c_str());
				retval += H(it->second->toString().c_str());
			}
		}

		return retval;
	}
};

}/*namespace __gnu_cxx*/
#endif /*WIN32*/

#endif /*PERSIDENTIFICATION_H_*/
