/**
 * @file result.h
 *
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */

#ifndef RESULT_H_
#define RESULT_H_

#include "resultBase.h"
#include "iopcmeta/typeDesc.h"
#include "simpleQuery.h"
#include "resultIterator.h"


namespace iopc {

template<typename T> class ResultIterator;

class SimpleQuery;


/**
 * @brief Executes a query and iterates through returned rows
 *
 * Result is used in the following way:
 * @code
 * Result<Person> queryResult(conn, query);
 * queryResult.open();
 * for (Result<Person>::iterator it = queryResult.begin(); it != queryResult.end(); it++) {
 *    cout << it->name << "   " << it->age << endl;
 * }
 * queryResult.close();
 * @endcode
 */
template<typename T>
class DLL_EXTERNAL Result : public ResultBase {

	friend class ResultIterator<T>;
private:
	DbPtr<T> currentObject;

	DbPtr<T> get(int i) {
		return DbPtr<T>(conn, getKeyAt(i));
	}

	DbPtr<T> getCurrent() {
		if (isTransient) {
			return currentObject;
		} else {
			return DbPtr<T>(conn, getCurrentKey());
		}
	}

	//~ResultBase();

protected:
	virtual bool loadCurrentObject(Cursor& cur) {
		currentObject = DbPtr<T>(cur);
		return currentObject.isValid();
	}


public:
	/**
	 * @brief Iterator type used to iterate through the query results
	 */
	typedef ResultIterator<T> iterator;

	/**
	 * @param conn Connection on which the query will be executed
	 * @param query The query to be executed
	 * @param preload If true, @a open() loads all keys (OIDs or KeyValues) and immediately releases the cursor it uses.
	 */
	Result(CachedConnection* conn, const Query& query, bool preload = true) : ResultBase(conn, query, preload, TypeDesc<T>::getType()) {};

	/**
	 * @brief Creates a @a Result that selects all persistent instances of the class T
	 * @param conn Connection on which the query will be executed
	 * @param preload If true, @a open() loads all keys (OIDs or KeyValues) and immediately releases the cursor it uses.
	 */
	Result(CachedConnection* conn, bool preload = true) : ResultBase(conn, SimpleQuery::All, preload, TypeDesc<T>::getType()) {};

	/**
	 * @brief Returns an @a ResultIterator<T> pointing to the first fetched object
	 */
	iterator begin() { return iterator(true, preload ? getKeyCount() : -1, *this); }

	/**
	 * @brief Returns an @a ResultIterator<T> representing an end of the fetched list.
	 */
	iterator end() { return iterator(false, preload ? getKeyCount() : -1, *this); }



};

}

#endif /*RESULT_H_*/
