/**
 * @file resultIterator.h
 *
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */

#ifndef RESULTITERATOR_H_
#define RESULTITERATOR_H_

#include "iopccommon/trace.h"
#include "iopccommon/exceptionBase.h"
#include "result.h"

namespace iopc {

template<typename T> class Result;

template<typename T>

/**
 * @brief An iterator that is used to iterate through query results represented as the @a Result<T> template class.
 *
 * @a ResultIterator overrides the -> operator to provide transparent access to the persistent object attributes.
 * Dereference operator (*) can be used to obtain a database pointer representing the persistent object.
 *
 * @ingroup iopclib
 */
class ResultIterator
{
	friend class Result<T>;
private:
	int pos;
	int count;
	const Type& type;
	Result<T>& results;
	bool begin;
	bool end;


	ResultIterator(bool begin, int count, Result<T>& results) : pos(begin ? 0 : count), count(count), type(TypeDesc<T>::getType()), results(results), begin(count == 0 || begin), end(count == 0 || !begin)  {};

	void Next()	{
		if (end) return;

		begin = false;
		pos++;

		if (count == -1) {
			// continuous fetching
			if (!results.fetchNext()) {
				end = true;
			}
		} else {
			// prefetched resultset
			if (pos == count) {
				end = true;
			}
		}
	}

	void Prev() {
		if (begin) return;

		end = false;
		pos--;

		if (count == -1) {
			// continuous fetching
			if (pos == 0) {
				begin = true;
			}
			if (!results.fetchPrev()) {
				IOPC_LOG_THROW(
					IopcInvalidStateException()
				);
			}
		} else {
			// prefetched resultset
			if (pos == 0) {
				begin = true;
			}
		}
	}

public:
	// Not perfect. May not work if comparing iterators from different queries.
	bool operator==(const ResultIterator<T>& it) {
		return
				count == it.count &&
				begin == it.begin &&
				end == it.end;
	}

	bool operator!=(const ResultIterator<T>& it) {
		return !(*this == it);
	}

	ResultIterator<T>& operator++() {
		Next();
		return *this;
	}

	ResultIterator<T>& operator++(int) {
		Next();
		return *this;
	}

	ResultIterator<T>& operator--() {
		Prev();
		return *this;
	}

	ResultIterator<T>& operator--(int) {
		Prev();
		return *this;
	}


	DbPtr<T> operator*() const {
		if (count == -1) {
			return results.getCurrent();
		} else {
			return results.get(pos);
		}
	}

	DbPtr<T> operator->() const {
		if (count == -1) {
			return results.getCurrent();
		} else {
			return results.get(pos);
		}
	}


};


}

#endif /*RESULTITERATOR_H_*/
